// Generated with `{{generated_with}}`
// From template at `{{generated_from}}`

use uv_pep440::{Prerelease, PrereleaseKind};
use crate::PythonVariant;
use crate::platform::ArchVariant;

pub(crate) const PYTHON_DOWNLOADS: &[ManagedPythonDownload] = &[
    {{#versions}}
    ManagedPythonDownload {
        key: PythonInstallationKey {
            major: {{value.major}},
            minor: {{value.minor}},
            patch: {{value.patch}},
            prerelease: {{value.prerelease}},
            implementation: LenientImplementationName::Known(ImplementationName::{{value.name}}),
            arch: Arch{
                family: target_lexicon::Architecture::{{value.arch_family}},
                variant: {{value.arch_variant}},
            },
            os: Os(target_lexicon::OperatingSystem::{{value.os}}),
            {{#value.libc}}
            libc: Libc::Some(target_lexicon::Environment::{{.}}),
            {{/value.libc}}
            {{^value.libc}}
            libc: Libc::None,
            {{/value.libc}}
            variant: {{value.variant}}
        },
        url: "{{value.url}}",
        {{#value.sha256}}
        sha256: Some("{{.}}")
        {{/value.sha256}}
        {{^value.sha256}}
        sha256: None
        {{/value.sha256}}
    },
    {{/versions}}
];
